from typing import Any, Sequence, Union

try:
    from polars.polars import when as pywhen

    _DOCUMENTING = False
except ImportError:  # pragma: no cover
    _DOCUMENTING = True

from polars import internals as pli


class WhenThenThen:
    """
    Utility class. See the `when` function.
    """

    def __init__(self, pywhenthenthen: Any):
        self.pywhenthenthen = pywhenthenthen

    def when(self, predicate: pli.Expr) -> "WhenThenThen":
        """
        Start another when, then, otherwise layer.
        """
        return WhenThenThen(self.pywhenthenthen.when(predicate._pyexpr))

    def then(
        self,
        expr: Union[
            pli.Expr,
            int,
            float,
            str,
            None,
            Sequence[
                Union[
                    int,
                    float,
                    str,
                    None,
                ]
            ],
        ],
    ) -> "WhenThenThen":
        """
        Values to return in case of the predicate being `True`.

        See Also: the `when` function.
        """
        expr_ = pli.expr_to_lit_or_expr(expr)
        return WhenThenThen(self.pywhenthenthen.then(expr_._pyexpr))

    def otherwise(
        self,
        expr: Union[
            pli.Expr,
            int,
            float,
            str,
            None,
            Sequence[
                Union[
                    int,
                    float,
                    str,
                    None,
                ]
            ],
        ],
    ) -> pli.Expr:
        """
        Values to return in case of the predicate being `False`.

        See Also: the `when` function.
        """
        expr = pli.expr_to_lit_or_expr(expr)
        return pli.wrap_expr(self.pywhenthenthen.otherwise(expr._pyexpr))


class WhenThen:
    """
    Utility class. See the `when` function.
    """

    def __init__(self, pywhenthen: Any):
        self._pywhenthen = pywhenthen

    def when(self, predicate: pli.Expr) -> WhenThenThen:
        """
        Start another when, then, otherwise layer.
        """
        return WhenThenThen(self._pywhenthen.when(predicate._pyexpr))

    def otherwise(self, expr: Union[pli.Expr, int, float, str]) -> pli.Expr:
        """
        Values to return in case of the predicate being `False`.

        See Also: the `when` function.
        """
        expr = pli.expr_to_lit_or_expr(expr)
        return pli.wrap_expr(self._pywhenthen.otherwise(expr._pyexpr))


class When:
    """
    Utility class. See the `when` function.
    """

    def __init__(self, pywhen: "pywhen"):
        self._pywhen = pywhen

    def then(
        self,
        expr: Union[
            pli.Expr, int, float, str, None, Sequence[Union[None, int, float, str]]
        ],
    ) -> WhenThen:
        """
        Values to return in case of the predicate being `True`.

        See Also: the `when` function.
        """
        expr = pli.expr_to_lit_or_expr(expr)
        pywhenthen = self._pywhen.then(expr._pyexpr)
        return WhenThen(pywhenthen)


def when(expr: pli.Expr) -> When:
    """
    Start a when, then, otherwise expression.

    Examples
    --------

    Below we add a column with the value 1, where column "foo" > 2 and the value -1 where it isn't.

    >>> df = pl.DataFrame({"foo": [1, 3, 4], "bar": [3, 4, 0]})
    >>> df.with_column(pl.when(pl.col("foo") > 2).then(pl.lit(1)).otherwise(pl.lit(-1)))
    shape: (3, 3)
    ┌─────┬─────┬─────────┐
    │ foo ┆ bar ┆ literal │
    │ --- ┆ --- ┆ ---     │
    │ i64 ┆ i64 ┆ i32     │
    ╞═════╪═════╪═════════╡
    │ 1   ┆ 3   ┆ -1      │
    ├╌╌╌╌╌┼╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
    │ 3   ┆ 4   ┆ 1       │
    ├╌╌╌╌╌┼╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
    │ 4   ┆ 0   ┆ 1       │
    └─────┴─────┴─────────┘

    Or with multiple `when, thens` chained:

    >>> df.with_column(
    ...     pl.when(pl.col("foo") > 2)
    ...     .then(1)
    ...     .when(pl.col("bar") > 2)
    ...     .then(4)
    ...     .otherwise(-1)
    ... )
    shape: (3, 3)
    ┌─────┬─────┬─────────┐
    │ foo ┆ bar ┆ literal │
    │ --- ┆ --- ┆ ---     │
    │ i64 ┆ i64 ┆ i32     │
    ╞═════╪═════╪═════════╡
    │ 1   ┆ 3   ┆ 4       │
    ├╌╌╌╌╌┼╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
    │ 3   ┆ 4   ┆ 1       │
    ├╌╌╌╌╌┼╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
    │ 4   ┆ 0   ┆ 1       │
    └─────┴─────┴─────────┘

    """
    expr = pli.expr_to_lit_or_expr(expr)
    pw = pywhen(expr._pyexpr)
    return When(pw)



import ctypes
import sys
from datetime import date, datetime, time, timedelta
from typing import Any, Dict, Optional, Sequence, Type, Union

try:
    import pyarrow as pa

    _PYARROW_AVAILABLE = True
except ImportError:  # pragma: no cover
    _PYARROW_AVAILABLE = False

from _ctypes import _SimpleCData  # type: ignore

try:
    from polars.polars import dtype_str_repr

    _DOCUMENTING = False
except ImportError:  # pragma: no cover
    _DOCUMENTING = True


class DataType:
    """Base class for all Polars data types"""

    @classmethod
    def string_repr(cls) -> str:
        return dtype_str_repr(cls)

    def __repr__(self) -> str:
        return dtype_str_repr(self)


PolarsDataType = Union[Type[DataType], DataType]


class Int8(DataType):
    """8-bit signed integer type"""

    pass


class Int16(DataType):
    """16-bit signed integer type"""

    pass


class Int32(DataType):
    """32-bit signed integer type"""

    pass


class Int64(DataType):
    """64-bit signed integer type"""

    pass


class UInt8(DataType):
    """8-bit unsigned integer type"""

    pass


class UInt16(DataType):
    """16-bit unsigned integer type"""

    pass


class UInt32(DataType):
    """32-bit unsigned integer type"""

    pass


class UInt64(DataType):
    """64-bit unsigned integer type"""

    pass


class Float32(DataType):
    """32-bit floating point type"""

    pass


class Float64(DataType):
    """64-bit floating point type"""

    pass


class Boolean(DataType):
    """Boolean type"""

    pass


class Utf8(DataType):
    """UTF-8 encoded string type"""

    pass


class Null(DataType):
    """Type representing Null / None values"""

    pass


class List(DataType):
    def __init__(self, inner: Type[DataType]):
        """
        Nested list/array type

        Parameters
        ----------
        inner
            The `DataType` of values within the list
        """
        self.inner = py_type_to_dtype(inner)

    def __eq__(self, other: Type[DataType]) -> bool:  # type: ignore
        # The comparison allows comparing objects to classes
        # and specific inner types to none specific.
        # if one of the arguments is not specific about its inner type
        # we infer it as being equal.
        # List[i64] == List[i64] == True
        # List[i64] == List == True
        # List[i64] == List[None] == True
        # List[i64] == List[f32] == False

        # allow comparing object instances to class
        if type(other) is type and issubclass(other, List):
            return True
        if isinstance(other, List):
            if self.inner is None or other.inner is None:
                return True
            else:
                return self.inner == other.inner
        else:
            return False

    def __hash__(self) -> int:
        return hash(List)


class Date(DataType):
    """Calendar date type"""

    pass


class Datetime(DataType):
    """Calendar date and time type"""

    def __init__(self, time_unit: str = "us", time_zone: Optional[str] = None):
        """
        Calendar date and time type

        Parameters
        ----------
        time_unit
            Any of {'ns', 'us', 'ms'}
        time_zone
            Timezone string as defined in pytz
        """
        self.tu = time_unit
        self.tz = time_zone

    def __eq__(self, other: Type[DataType]) -> bool:  # type: ignore
        # allow comparing object instances to class
        if type(other) is type and issubclass(other, Datetime):
            return True
        if isinstance(other, Datetime):
            return self.tu == other.tu and self.tz == other.tz
        else:
            return False

    def __hash__(self) -> int:
        return hash(Datetime)


class Duration(DataType):
    """Time duration/delta type"""

    def __init__(self, time_unit: str = "us"):
        """
        Time duration/delta type

        Parameters
        ----------
        time_unit
            Any of {'ns', 'us', 'ms'}
        """
        self.tu = time_unit

    def __eq__(self, other: Type[DataType]) -> bool:  # type: ignore
        # allow comparing object instances to class
        if type(other) is type and issubclass(other, Duration):
            return True
        if isinstance(other, Duration):
            return self.tu == other.tu
        else:
            return False

    def __hash__(self) -> int:
        return hash(Duration)


class Time(DataType):
    """Time of day type"""

    pass


class Object(DataType):
    """Type for wrapping arbitrary Python objects"""

    pass


class Categorical(DataType):
    """A categorical encoding of a set of strings"""

    pass


class Field:
    def __init__(self, name: str, dtype: Type[DataType]):
        """
        Definition of a single field within a `Struct` DataType

        Parameters
        ----------
        name
            The name of the field within its parent `Struct`
        dtype
            The `DataType` of the field's values
        """
        self.name = name
        self.dtype = py_type_to_dtype(dtype)

    def __eq__(self, other: "Field") -> bool:  # type: ignore
        return (self.name == other.name) & (self.dtype == other.dtype)

    def __repr__(self) -> str:
        if isinstance(self.dtype, type):
            dtype_str = self.dtype.string_repr()
        else:
            dtype_str = repr(self.dtype)
        return f'Field("{self.name}": {dtype_str})'


class Struct(DataType):
    def __init__(self, fields: Sequence[Field]):
        """
        Struct composite type

        Parameters
        ----------
        fields
            The sequence of fields that make up the struct
        """
        self.fields = fields

    def __eq__(self, other: Type[DataType]) -> bool:  # type: ignore
        # The comparison allows comparing objects to classes
        # and specific inner types to none specific.
        # if one of the arguments is not specific about its inner type
        # we infer it as being equal.
        # See the list type for more info
        if type(other) is type and issubclass(other, Struct):
            return True
        if isinstance(other, Struct):
            if self.fields is None or other.fields is None:
                return True
            else:
                return self.fields == other.fields
        else:
            return False

    def __hash__(self) -> int:
        return hash(Struct)


_DTYPE_TO_FFINAME: Dict[PolarsDataType, str] = {
    Int8: "i8",
    Int16: "i16",
    Int32: "i32",
    Int64: "i64",
    UInt8: "u8",
    UInt16: "u16",
    UInt32: "u32",
    UInt64: "u64",
    Float32: "f32",
    Float64: "f64",
    Boolean: "bool",
    Utf8: "str",
    List: "list",
    Date: "date",
    Datetime: "datetime",
    Duration: "duration",
    Time: "time",
    Object: "object",
    Categorical: "categorical",
    Struct: "struct",
}

_DTYPE_TO_CTYPE: Dict[PolarsDataType, Any] = {
    UInt8: ctypes.c_uint8,
    UInt16: ctypes.c_uint16,
    UInt32: ctypes.c_uint32,
    UInt64: ctypes.c_uint64,
    Int8: ctypes.c_int8,
    Int16: ctypes.c_int16,
    Int32: ctypes.c_int32,
    Date: ctypes.c_int32,
    Int64: ctypes.c_int64,
    Float32: ctypes.c_float,
    Float64: ctypes.c_double,
    Datetime: ctypes.c_int64,
    Duration: ctypes.c_int64,
    Time: ctypes.c_int64,
}


_PY_TYPE_TO_DTYPE: Dict[type, Type[DataType]] = {
    float: Float64,
    int: Int64,
    str: Utf8,
    bool: Boolean,
    date: Date,
    datetime: Datetime,
    timedelta: Duration,
    time: Time,
    list: List,
    tuple: List,
}


_DTYPE_TO_PY_TYPE: Dict[PolarsDataType, type] = {
    Float64: float,
    Float32: float,
    Int64: int,
    Int32: int,
    Int16: int,
    Int8: int,
    Utf8: str,
    UInt8: int,
    UInt16: int,
    UInt32: int,
    UInt64: int,
    Boolean: bool,
    Duration: timedelta,
    Datetime: datetime,
    Date: date,
    Time: time,
}

# Map Numpy char codes to polars dtypes.
#
# Windows behaves differently from other platforms as C long is
# only 32-bit on Windows, while it is 64-bit on other platforms.
# See: https://numpy.org/doc/stable/reference/arrays.scalars.html
_NUMPY_CHAR_CODE_TO_DTYPE = {
    "b": Int8,
    "h": Int16,
    "i": Int32,
    ("q" if sys.platform == "win32" else "l"): Int64,
    "B": UInt8,
    "H": UInt16,
    "I": UInt32,
    ("Q" if sys.platform == "win32" else "L"): UInt64,
    "f": Float32,
    "d": Float64,
    "?": Boolean,
}

if _PYARROW_AVAILABLE:
    _PY_TYPE_TO_ARROW_TYPE: Dict[type, "pa.lib.DataType"] = {
        float: pa.float64(),
        int: pa.int64(),
        str: pa.large_utf8(),
        bool: pa.bool_(),
        date: pa.date32(),
        time: pa.time64("us"),
        datetime: pa.timestamp("us"),
        timedelta: pa.duration("us"),
    }

    _DTYPE_TO_ARROW_TYPE = {
        Int8: pa.int8(),
        Int16: pa.int16(),
        Int32: pa.int32(),
        Int64: pa.int64(),
        UInt8: pa.uint8(),
        UInt16: pa.uint16(),
        UInt32: pa.uint32(),
        UInt64: pa.uint64(),
        Float32: pa.float32(),
        Float64: pa.float64(),
        Boolean: pa.bool_(),
        Utf8: pa.large_utf8(),
        Date: pa.date32(),
        # handle temporal types that require units
        Datetime: pa.timestamp("us"),
        (Datetime, "ms"): pa.timestamp("ms"),
        (Datetime, "us"): pa.timestamp("us"),
        (Datetime, "ns"): pa.timestamp("ns"),
        Duration: pa.duration("us"),
        (Duration, "ms"): pa.duration("ms"),
        (Duration, "us"): pa.duration("us"),
        (Duration, "ns"): pa.duration("ns"),
        Time: pa.time64("us"),
        (Time, "ms"): pa.time32("ms"),
        (Time, "us"): pa.time64("us"),
        (Time, "ns"): pa.time64("ns"),
    }


def dtype_to_ctype(dtype: PolarsDataType) -> Type[_SimpleCData]:
    try:
        return _DTYPE_TO_CTYPE[dtype]
    except KeyError:  # pragma: no cover
        raise NotImplementedError


def dtype_to_ffiname(dtype: PolarsDataType) -> str:
    try:
        return _DTYPE_TO_FFINAME[dtype]
    except KeyError:  # pragma: no cover
        raise NotImplementedError


def dtype_to_py_type(dtype: PolarsDataType) -> Type:
    try:
        return _DTYPE_TO_PY_TYPE[dtype]
    except KeyError:  # pragma: no cover
        raise NotImplementedError


def is_polars_dtype(data_type: Any) -> bool:
    return (
        type(data_type) is type
        and issubclass(data_type, DataType)
        or isinstance(data_type, DataType)
    )


def py_type_to_dtype(data_type: Any) -> Type[DataType]:
    # when the passed in is already a Polars datatype, return that
    if is_polars_dtype(data_type):
        return data_type
    try:
        return _PY_TYPE_TO_DTYPE[data_type]
    except KeyError:  # pragma: no cover
        raise NotImplementedError


def py_type_to_arrow_type(dtype: Type[Any]) -> "pa.lib.DataType":
    """
    Convert a Python dtype to an Arrow dtype.
    """
    try:
        return _PY_TYPE_TO_ARROW_TYPE[dtype]
    except KeyError:  # pragma: no cover
        raise ValueError(f"Cannot parse dtype {dtype} into Arrow dtype.")


def dtype_to_arrow_type(dtype: PolarsDataType) -> "pa.lib.DataType":
    """
    Convert a Polars dtype to an Arrow dtype.
    """
    try:
        unit = getattr(dtype, "tu", None)
        lookup = dtype if unit is None else (dtype, unit)
        return _DTYPE_TO_ARROW_TYPE[lookup]
    except KeyError:  # pragma: no cover
        raise ValueError(f"Cannot parse dtype {dtype} into Arrow dtype.")


def supported_numpy_char_code(dtype: str) -> bool:
    return dtype in _NUMPY_CHAR_CODE_TO_DTYPE


def numpy_char_code_to_dtype(dtype: str) -> Type[DataType]:
    try:
        return _NUMPY_CHAR_CODE_TO_DTYPE[dtype]
    except KeyError:  # pragma: no cover
        raise NotImplementedError


def maybe_cast(
    el: Type[DataType], dtype: Type, time_unit: Optional[str] = None
) -> Type[DataType]:
    # cast el if it doesn't match
    from polars.utils import _datetime_to_pl_timestamp, _timedelta_to_pl_timedelta

    if isinstance(el, datetime):
        return _datetime_to_pl_timestamp(el, time_unit)
    elif isinstance(el, timedelta):
        return _timedelta_to_pl_timedelta(el, time_unit)
    py_type = dtype_to_py_type(dtype)
    if not isinstance(el, py_type):
        el = py_type(el)
    return el


#: Mapping of `~polars.DataFrame` / `~polars.LazyFrame` column names to their `DataType`
Schema = Dict[str, Type[DataType]]



try:
    from polars.polars import (
        ArrowError,
        ComputeError,
        DuplicateError,
        NoDataError,
        NotFoundError,
        PanicException,
        SchemaError,
        ShapeError,
    )
except ImportError:  # pragma: no cover
    # They are only redefined for documentation purposes
    # when there is no binary yet

    class ArrowError(Exception):  # type: ignore
        """Exception raised the underlying Arrow library encounters an error"""

        pass

    class ComputeError(Exception):  # type: ignore
        """Exception raised when we couldn't finish the computation"""

        pass

    class NoDataError(Exception):  # type: ignore
        """Exception raised when an operation can not be performed on an empty data structure"""

        pass

    class NotFoundError(Exception):  # type: ignore
        """Exception raised when a specified column is not found"""

        pass

    class SchemaError(Exception):  # type: ignore
        """Exception raised when trying to combine data structures with mismatched schemas"""

        pass

    class ShapeError(Exception):  # type: ignore
        """Exception raised when trying to combine data structures with incompatible shapes"""

        pass

    class DuplicateError(Exception):  # type: ignore
        """Exception raised when a column name is duplicated"""

        pass

    class PanicException(Exception):  # type: ignore
        """Exception raised when an unexpected state causes a panic in the underlying Rust library"""

        pass


__all__ = [
    "ArrowError",
    "ComputeError",
    "NoDataError",
    "NotFoundError",
    "SchemaError",
    "ShapeError",
    "DuplicateError",
    "PanicException",
]



from typing import Any

from polars.datatypes import (
    Boolean,
    Float32,
    Float64,
    Int16,
    Int32,
    Int64,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
    Utf8,
    dtype_to_py_type,
)
from polars.internals import DataFrame, Series

_NUMERIC_COL_TYPES = (
    Int16,
    Int32,
    Int64,
    UInt16,
    UInt32,
    UInt64,
    UInt8,
    Utf8,
    Float32,
    Float64,
    Boolean,
)


def assert_frame_equal(
    left: DataFrame,
    right: DataFrame,
    check_dtype: bool = True,
    check_exact: bool = False,
    check_column_names: bool = True,
    rtol: float = 1.0e-5,
    atol: float = 1.0e-8,
) -> None:
    """
    Raise detailed AssertionError if `left` does not equal `right`.


    Parameters
    ----------
    left
        the dataframe to compare
    right
        the dataframe to compare with
    check_dtype
        if True, data types need to match exactly
    check_exact
        if False, test if values are within tolerance of each other (see `rtol` & `atol`)
    check_column_names
        if True, dataframes must have the same column names in the same order
    rtol
        relative tolerance for inexact checking. Fraction of values in `right`
    atol
        absolute tolerance for inexact checking.

    Returns
    -------

    Examples
    --------

    >>> df1 = pl.DataFrame({"a": [1, 2, 3]})
    >>> df2 = pl.DataFrame({"a": [2, 3, 4]})
    >>> pl.testing.assert_frame_equal(df1, df2)  # doctest: +SKIP
    """

    obj = "DataFrame"

    if not (isinstance(left, DataFrame) and isinstance(right, DataFrame)):
        raise_assert_detail(obj, "Type mismatch", type(left), type(right))

    if left.shape[0] != right.shape[0]:
        raise_assert_detail(obj, "Length mismatch", left.shape, right.shape)

    # this assumes we want it in the same order
    union_cols = list(set(left.columns).union(set(right.columns)))
    for c in union_cols:
        if c not in right.columns:
            raise AssertionError(
                f"column {c} in left dataframe, but not in right dataframe"
            )
        if c not in left.columns:
            raise AssertionError(
                f"column {c} in right dataframe, but not in left dataframe"
            )

    if check_column_names:
        if left.columns != right.columns:
            raise AssertionError("Columns are not in the same order")

    # this does not assume a particular order
    for col in left.columns:
        _assert_series_inner(
            left[col], right[col], check_dtype, check_exact, atol, rtol, obj
        )


def assert_series_equal(
    left: Series,
    right: Series,
    check_dtype: bool = True,
    check_names: bool = True,
    check_exact: bool = False,
    rtol: float = 1.0e-5,
    atol: float = 1.0e-8,
) -> None:
    """
    Raise detailed AssertionError if `left` does not equal `right`.

    Parameters
    ----------
    left
        the series to compare
    right
        the series to compare with
    check_dtype
        if True, data types need to match exactly
    check_names
        if True, names need to match
    check_exact
        if False, test if values are within tolerance of each other (see `rtol` & `atol`)
    rtol
        relative tolerance for inexact checking. Fraction of values in `right`
    atol
        absolute tolerance for inexact checking.

    Returns
    -------

    Examples
    --------

    >>> s1 = pl.Series([1, 2, 3])
    >>> s2 = pl.Series([2, 3, 4])
    >>> pl.testing.assert_series_equal(s1, s2)  # doctest: +SKIP
    """
    obj = "Series"

    if not (isinstance(left, Series) and isinstance(right, Series)):
        raise_assert_detail(obj, "Type mismatch", type(left), type(right))

    if left.shape != right.shape:
        raise_assert_detail(obj, "Shape mismatch", left.shape, right.shape)

    if check_names:
        if left.name != right.name:
            raise_assert_detail(obj, "Name mismatch", left.name, right.name)

    _assert_series_inner(left, right, check_dtype, check_exact, atol, rtol, obj)


def _assert_series_inner(
    left: Series,
    right: Series,
    check_dtype: bool,
    check_exact: bool,
    atol: float,
    rtol: float,
    obj: str,
) -> None:
    """
    Compares Series dtype + values
    """
    try:
        can_be_subtracted = hasattr(dtype_to_py_type(left.dtype), "__sub__")
    except NotImplementedError:
        can_be_subtracted = False

    check_exact = check_exact or not can_be_subtracted or left.dtype == Boolean

    if check_dtype:
        if left.dtype != right.dtype:
            raise_assert_detail(obj, "Dtype mismatch", left.dtype, right.dtype)

    if check_exact:
        if (left != right).sum() != 0:
            raise_assert_detail(
                obj, "Exact value mismatch", left=list(left), right=list(right)
            )
    else:
        if ((left - right).abs() > (atol + rtol * right.abs())).sum() != 0:
            raise_assert_detail(
                obj, "Value mismatch", left=list(left), right=list(right)
            )


def raise_assert_detail(
    obj: str,
    message: str,
    left: Any,
    right: Any,
) -> None:
    __tracebackhide__ = True

    msg = f"""{obj} are different

{message}"""

    msg += f"""
[left]:  {left}
[right]: {right}"""

    raise AssertionError(msg)



from types import TracebackType
from typing import Optional, Type

try:
    from polars.polars import toggle_string_cache as pytoggle_string_cache

    _DOCUMENTING = False
except ImportError:  # pragma: no cover
    _DOCUMENTING = True


class StringCache:
    """
    Context manager that allows data sources to share the same categorical features.
    This will temporarily cache the string categories until the context manager is finished.

    >>> df = pl.DataFrame(
    ...     {
    ...         "a_col": ["red", "green", "blue"],
    ...         "b_col": ["yellow", "orange", "black"],
    ...     }
    ... )
    >>> with pl.StringCache():
    ...     df = df.with_columns(
    ...         [
    ...             pl.col("a_col").cast(pl.Categorical).alias("a_col"),
    ...             pl.col("b_col").cast(pl.Categorical).alias("b_col"),
    ...         ]
    ...     )
    ...
    """

    def __init__(self) -> None:
        pass

    def __enter__(self) -> "StringCache":
        pytoggle_string_cache(True)
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        pytoggle_string_cache(False)


def toggle_string_cache(toggle: bool) -> None:
    """
    Turn on/off the global string cache. This ensures that casts to Categorical types have the categories when string
    values are equal.
    """
    pytoggle_string_cache(toggle)
